"""
This module contains the logic to create the compatibility matrix from a list of changes.
"""

import csv
import itertools
from enum import StrEnum
from pathlib import Path
from typing import Any as _Any
from typing import Mapping, Sequence

from . import change_schemas


class ChangeSymbol(StrEnum):
    """
    This enum class lists the different symbols of changes in the compatibility matrix.
    """

    CHANGE_NONE = "🟢"
    CHANGE_NON_CRITICAL = "🟡"
    CHANGE_CRITICAL = "🔴"
    NON_EXISTENT = "\\-"
    ADDED = "➕"
    REMOVED = "➖"


def determine_symbol(
    changes: Sequence[change_schemas.Change], namespace: Sequence[tuple[str, ...]], cls: tuple[str, ...]
) -> ChangeSymbol:
    """
    Determine the symbol of a change.
    """
    if len(changes) == 1 and changes[0].type == change_schemas.ChangeType.CLASS_REMOVED:
        return ChangeSymbol.REMOVED
    if len(changes) == 1 and changes[0].type == change_schemas.ChangeType.CLASS_ADDED:
        return ChangeSymbol.ADDED
    if cls not in namespace:
        return ChangeSymbol.NON_EXISTENT
    if len(changes) == 0:
        return ChangeSymbol.CHANGE_NONE

    assert all(
        change.type not in (change_schemas.ChangeType.CLASS_ADDED, change_schemas.ChangeType.CLASS_REMOVED)
        for change in changes
    ), "Internal error: CLASS_ADDED and CLASS_REMOVED must be the only change per class if present."
    if any(change_schemas.is_change_critical(change) for change in changes):
        return ChangeSymbol.CHANGE_CRITICAL
    return ChangeSymbol.CHANGE_NON_CRITICAL


def create_compatibility_matrix_csv(
    output: Path,
    versions: Sequence[str],
    namespaces: Mapping[str, Sequence[tuple[str, ...]]],
    changes: Mapping[tuple[str, str], Sequence[change_schemas.Change]],
) -> None:
    """
    Create a compatibility matrix csv file from the given changes.
    """
    output.parent.mkdir(parents=True, exist_ok=True)
    with open(output, "w", encoding="utf-8") as file:
        csv_writer = csv.writer(file, delimiter=",", lineterminator="\n", escapechar="/")
        csv_writer.writerow(("", *versions[1:]))
        all_classes: set[tuple[str, ...]] = set(itertools.chain.from_iterable(namespaces.values()))

        for class_path in sorted(all_classes, key=lambda cls: tuple(cls_part.lower() for cls_part in cls)):
            row = [class_path[-1]]
            class_path_str = "/".join(class_path) + "#"
            for version_old, version_new in itertools.pairwise(versions):
                changes_related_to_class = [
                    change
                    for change in changes[(version_old, version_new)]
                    if change.old_trace.startswith(class_path_str) or change.new_trace.startswith(class_path_str)
                ]
                row.append(determine_symbol(changes_related_to_class, namespaces[version_new], class_path).value)
            csv_writer.writerow(row)



"""
Contains AufAbschlagstaffelProOrt class
"""

from decimal import Decimal
from typing import Annotated, Literal, Optional

from pydantic import Field

from ..enum.comtyp import ComTyp
from ..utils import postprocess_docstring
from .com import COM

# pylint: disable=too-few-public-methods


@postprocess_docstring
class AufAbschlagstaffelProOrt(COM):
    """
    Gibt den Wert eines Auf- oder Abschlags und dessen Staffelgrenzen an

    .. raw:: html

        <object data="../_static/images/bo4e/com/AufAbschlagstaffelProOrt.svg" type="image/svg+xml"></object>

    .. HINT::
        `AufAbschlagstaffelProOrt JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/com/AufAbschlagstaffelProOrt.json>`_

    """

    typ: Annotated[Literal[ComTyp.AUFABSCHLAGSTAFFELPROORT], Field(alias="_typ")] = ComTyp.AUFABSCHLAGSTAFFELPROORT

    wert: Optional[Decimal] = None
    """Der Wert für den Auf- oder Abschlag."""
    staffelgrenze_von: Optional[Decimal] = None
    """Unterer Wert, ab dem die Staffel gilt."""
    staffelgrenze_bis: Optional[Decimal] = None
    """Oberer Wert, bis zu dem die Staffel gilt."""



"""
This module provides functions to compare the BO4E JSON schemas of different versions.
It also contains functions to query GitHub for the latest BO4E versions to compare with the schemas of the current
work tree.
Additionally, it implements a little cache functionality to avoid multiple downloads of the same versions e.g.
if you're testing locally.
"""

import logging
import re
from pathlib import Path

import bost.operations
from bost.pull import OWNER, REPO

from . import diff, loader, matrix, versioning

logger = logging.getLogger(__name__)

REGEX_RELEASE_VERSION = re.compile(r"^v(\d{6}\.\d+\.\d+)$")
REGEX_RELEASE_CANDIDATE_VERSION = re.compile(r"^v(\d{6}\.\d+\.\d+)-rc\d+$")


def _monkey_patch_bost_regex_if_local_testing(version: str) -> None:
    regex_expected_version = re.compile(r"^v\d+\.\d+\.\d+(?:-rc\d+)?$")
    if not regex_expected_version.fullmatch(version):
        bost.operations.REF_ONLINE_REGEX = re.compile(
            rf"^https://raw\.githubusercontent\.com/(?:{OWNER.upper()}|{OWNER.lower()}|Hochfrequenz)/{REPO}/"
            rf"(?P<version>[^/]+)/"
            r"src/bo4e_schemas/(?P<sub_path>(?:\w+/)*)(?P<model>\w+)\.json#?$"
        )


def create_tables_for_doc(
    compatibility_matrix_output_file: Path,
    gh_version: str,
    *,
    gh_token: str | None = None,
    last_n_versions: int = 2,
) -> None:
    """
    Creates the compatibility matrix for the documentation. The output is a csv file. This can be referenced
    inside Sphinx documentation. See https://sublime-and-sphinx-guide.readthedocs.io/en/latest/tables.html#csv-files
    for more information.
    If you have problems with rate limiting, please set gh_token.
    The compatibility matrix will be built for last_n_versions + the current version in the checkout working directory.
    If you set last_n_versions = 0 all versions since v202401.0.0 will be compared.
    Note: The matrix will never contain the first version as column. Each column is a comparison to the version before.
    Note: Only functional releases will be compared since technical releases are enforced to be fully compatible.
    See https://github.com/bo4e/BO4E-python/issues/784
    """
    _monkey_patch_bost_regex_if_local_testing(gh_version)
    logger.info("Retrieving the last %d release versions", last_n_versions)
    versions = list(
        reversed(list(versioning.get_last_n_tags(last_n_versions, ref=gh_version, exclude_technical_bumps=True)))
    )
    logger.info("Comparing versions iteratively: %s", " -> ".join([*versions, gh_version]))
    changes_iterables = diff.compare_bo4e_versions_iteratively(versions, gh_version, gh_token=gh_token)
    logger.info("Building namespaces")
    changes = {key: list(value) for key, value in changes_iterables.items()}
    namespaces = {version: list(loader.get_namespace(loader.BO4E_BASE_DIR / version)) for version in versions}
    namespaces[gh_version] = list(loader.get_namespace(loader.BO4E_BASE_DIR / gh_version))
    logger.info("Creating compatibility matrix")
    matrix.create_compatibility_matrix_csv(
        compatibility_matrix_output_file, [*versions, gh_version], namespaces, changes
    )


def test_create_tables_for_doc() -> None:
    """
    Test the create_tables_for_doc function locally without building the entire documentation.
    Needs the JSON schemas to be present in /json_schemas with TARGET_VERSION set to "local".
    """
    create_tables_for_doc(
        Path(__file__).parents[1] / "compatibility_matrix.csv",
        "local",
        last_n_versions=0,
    )



"""
Contains the classes to model changes between two BO4E versions.
"""

from enum import StrEnum
from typing import Any, Iterable

from pydantic import BaseModel


class ChangeType(StrEnum):
    """
    This enum class lists the different types of changes of a single change between two BO4E versions.
    """

    FIELD_ADDED = "field_added"
    FIELD_REMOVED = "field_removed"
    FIELD_DEFAULT_CHANGED = "field_default_changed"
    FIELD_DESCRIPTION_CHANGED = "field_description_changed"
    FIELD_TITLE_CHANGED = "field_title_changed"
    # field type change types
    FIELD_CARDINALITY_CHANGED = "field_cardinality_changed"
    FIELD_REFERENCE_CHANGED = "field_reference_changed"
    FIELD_STRING_FORMAT_CHANGED = "field_string_format_changed"
    FIELD_ANY_OF_TYPE_ADDED = "field_any_of_type_added"
    FIELD_ANY_OF_TYPE_REMOVED = "field_any_of_type_removed"
    FIELD_ALL_OF_TYPE_ADDED = "field_all_of_type_added"
    FIELD_ALL_OF_TYPE_REMOVED = "field_all_of_type_removed"
    FIELD_TYPE_CHANGED = "field_type_changed"  # An arbitrary unclassified change in type

    CLASS_ADDED = "class_added"
    CLASS_REMOVED = "class_removed"
    CLASS_DESCRIPTION_CHANGED = "class_description_changed"

    ENUM_VALUE_ADDED = "enum_value_added"
    ENUM_VALUE_REMOVED = "enum_value_removed"


class Change(BaseModel):
    """
    This pydantic class models a single change between two BO4E versions.
    """

    type: ChangeType
    old: Any
    new: Any
    old_trace: str
    new_trace: str

    def __str__(self) -> str:
        return f"{self.type}: {self.old} -> {self.new}"


def is_change_critical(change: Change) -> bool:
    """
    This function checks if a change is critical i.e. if the new value is incompatible to the old value.
    """
    return change.type in (
        ChangeType.FIELD_REMOVED,
        ChangeType.FIELD_TYPE_CHANGED,
        ChangeType.FIELD_CARDINALITY_CHANGED,
        ChangeType.FIELD_REFERENCE_CHANGED,
        ChangeType.FIELD_STRING_FORMAT_CHANGED,
        ChangeType.FIELD_ANY_OF_TYPE_ADDED,
        ChangeType.FIELD_ANY_OF_TYPE_REMOVED,
        ChangeType.FIELD_ALL_OF_TYPE_ADDED,
        ChangeType.FIELD_ALL_OF_TYPE_REMOVED,
        ChangeType.CLASS_REMOVED,
        ChangeType.ENUM_VALUE_REMOVED,
    )


def filter_non_crit(changes: Iterable[Change]) -> Iterable[Change]:
    """
    This function filters out all non-critical changes.
    """
    return (change for change in changes if is_change_critical(change))



"""
Contains Vertragskonditionen class
"""

from decimal import Decimal
from typing import TYPE_CHECKING, Annotated, Literal, Optional

from pydantic import Field

from ..enum.comtyp import ComTyp
from ..utils import postprocess_docstring
from .com import COM

if TYPE_CHECKING:

    from .zeitraum import Zeitraum

# pylint: disable=too-few-public-methods


@postprocess_docstring
class Vertragskonditionen(COM):
    """
    Abbildung für Vertragskonditionen. Die Komponente wird sowohl im Vertrag als auch im Tarif verwendet.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Vertragskonditionen.svg" type="image/svg+xml"></object>

    .. HINT::
        `Vertragskonditionen JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/com/Vertragskonditionen.json>`_

    """

    typ: Annotated[Literal[ComTyp.VERTRAGSKONDITIONEN], Field(alias="_typ")] = ComTyp.VERTRAGSKONDITIONEN

    beschreibung: Optional[str] = None
    """
    Freitext zur Beschreibung der Konditionen, z.B. "Standardkonditionen Gas"
    """
    anzahl_abschlaege: Optional[Decimal] = None
    """Anzahl der vereinbarten Abschläge pro Jahr, z.B. 12"""
    vertragslaufzeit: Optional["Zeitraum"] = None
    """Über diesen Zeitraum läuft der Vertrag"""
    kuendigungsfrist: Optional["Zeitraum"] = None
    """Innerhalb dieser Frist kann der Vertrag gekündigt werden"""
    vertragsverlaengerung: Optional["Zeitraum"] = None
    """Falls der Vertrag nicht gekündigt wird, verlängert er sich automatisch um die hier angegebene Zeit"""
    abschlagszyklus: Optional["Zeitraum"] = None
    """In diesen Zyklen werden Abschläge gestellt. Alternativ kann auch die Anzahl in den Konditionen angeben werden."""



"""
Contains Menge class
"""

from decimal import Decimal
from typing import TYPE_CHECKING, Annotated, Literal, Optional

from pydantic import Field

from ..enum.comtyp import ComTyp
from ..utils import postprocess_docstring
from .com import COM

if TYPE_CHECKING:
    from ..enum.mengeneinheit import Mengeneinheit


# pylint: disable=too-few-public-methods


@postprocess_docstring
class Menge(COM):
    """
    Abbildung einer Menge mit Wert und Einheit.

    .. raw:: html

        <object data="../_static/images/bo4e/com/Menge.svg" type="image/svg+xml"></object>

    .. HINT::
        `Menge JSON Schema <https://json-schema.app/view/%23?url=https://raw.githubusercontent.com/BO4E/BO4E-Schemas/{__gh_version__}/src/bo4e_schemas/com/Menge.json>`_

    """

    typ: Annotated[Literal[ComTyp.MENGE], Field(alias="_typ")] = ComTyp.MENGE

    wert: Optional[Decimal] = None
    """Gibt den absoluten Wert der Menge an"""
    einheit: Optional["Mengeneinheit"] = None
    """Gibt die Einheit zum jeweiligen Wert an"""



"""
Contains functions to load and save schema files and changes
"""

import json
import shutil
from pathlib import Path
from typing import Iterable

from bost import main as bost_main
from bost.operations import update_references as bost_update_references
from bost.pull import SchemaMetadata
from bost.schema import SchemaRootType
from pydantic import TypeAdapter

from . import change_schemas

BO4E_BASE_DIR = Path(__file__).parents[2] / "tmp/bo4e_json_schemas"
LOCAL_JSON_SCHEMA_DIR = Path(__file__).parents[2] / "json_schemas"


def load_schema_file(path: Path) -> SchemaRootType:
    """
    Load a schema file and return the parsed schema
    """
    return TypeAdapter(SchemaRootType).validate_json(path.read_text("utf-8"))
    # mypy has problems to infer the Union type here.


def load_changes(path: Path) -> list[change_schemas.Change]:
    """
    Load a changes file and return the parsed changes
    """
    return TypeAdapter(list[change_schemas.Change]).validate_json(path.read_text("utf-8"))


def save_changes(path: Path, changes: Iterable[change_schemas.Change]) -> None:
    """
    Save the changes to a file
    """
    with open(path, "w", encoding="utf-8") as file:
        json.dump(
            TypeAdapter(list[change_schemas.Change]).dump_python(list(changes), mode="json"),
            file,
        )


def get_namespace(path: Path) -> Iterable[tuple[str, ...]]:
    """
    Get the namespace from a file
    """
    for schema_file in path.rglob("*.json"):
        sub_path = schema_file.relative_to(path).parts[:-1]
        yield *sub_path, schema_file.stem


def pull_bo4e_version(version: str, output: Path, gh_token: str | None = None) -> None:
    """
    Pull the BO4E version from the given version string.
    """
    bost_main(
        output=output,
        target_version=version,
        update_refs=True,
        set_default_version=False,
        clear_output=True,
        token=gh_token,
    )


def update_references(path: Path, version: str) -> None:
    """
    Update the references in the given path. This step is needed for the local build.
    """
    schema_namespace = {}
    for schema_path in get_namespace(path):
        local_path = Path(path, *schema_path).with_suffix(".json")
        schema_namespace[schema_path[-1]] = SchemaMetadata(
            class_name=schema_path[-1],
            download_url="",
            module_path=schema_path,
            file_path=local_path,
            cached_path=local_path,
            token=None,
        )
    for schema_metadata in schema_namespace.values():
        bost_update_references(schema_metadata, schema_namespace, version)
        schema_metadata.save()


def pull_or_reuse_bo4e_version(version: str, gh_token: str | None = None, from_local: bool = False) -> Path:
    """
    Pull the BO4E version from the given version string or reuse the version if it was already pulled before.
    If version is None use the BO4E version of the checkout working directory by assuming the compiled json
    schemas in /json_schemas.
    Returns the path of the bo4e directory.
    """
    bo4e_dir = BO4E_BASE_DIR / version

    if from_local:
        if not any(LOCAL_JSON_SCHEMA_DIR.rglob("*.json")):
            raise ValueError(
                "No local json schemas found in /json_schemas. "
                "Please ensure that the json schemas are build on beforehand."
            )
        if bo4e_dir.exists():
            shutil.rmtree(bo4e_dir)
        shutil.copytree(LOCAL_JSON_SCHEMA_DIR, bo4e_dir)
        update_references(bo4e_dir, version)
    elif any(bo4e_dir.rglob("*.json")):
        return bo4e_dir
    else:
        pull_bo4e_version(version, bo4e_dir, gh_token)
    return bo4e_dir



"""
This module provides a CLI to check if a version tag has the expected format we expect in the BO4E repository.
"""

import functools
import logging
import re
import subprocess
import sys
from typing import ClassVar, Iterable, Literal, Optional

import click
from github import Github
from github.Auth import Token
from github.Repository import Repository
from more_itertools import one
from pydantic import BaseModel, ConfigDict

from . import diff

logger = logging.getLogger(__name__)


@functools.total_ordering
class Version(BaseModel):
    """
    A class to represent a BO4E version number.
    """

    version_pattern: ClassVar[re.Pattern[str]] = re.compile(
        r"^v(?P<major>\d{6})\.(?P<functional>\d+)\.(?P<technical>\d+)(?:-rc(?P<candidate>\d+))?$"
    )

    major: int
    functional: int
    technical: int
    candidate: Optional[int] = None
    model_config = ConfigDict(frozen=True)

    @classmethod
    def from_string(cls, version: str, allow_candidate: bool = False) -> "Version":
        """
        Parse a version string and return a Version object.
        Raises a ValueError if the version string does not match the expected pattern.
        Raises a ValueError if allow_candidate is False and the version string contains a candidate version.
        """
        match = cls.version_pattern.fullmatch(version)
        if match is None:
            raise ValueError(f"Expected version to match {cls.version_pattern}, got {version}")
        inst = cls(
            major=int(match.group("major")),
            functional=int(match.group("functional")),
            technical=int(match.group("technical")),
            candidate=int(match.group("candidate")) if match.group("candidate") is not None else None,
        )
        if not allow_candidate and inst.is_candidate():
            raise ValueError(f"Expected a version without candidate, got a candidate version: {version}")
        return inst

    @property
    def tag_name(self) -> str:
        """
        Return the tag name for this version.
        """
        return f"v{self.major}.{self.functional}.{self.technical}" + (
            f"-rc{self.candidate}" if self.is_candidate() else ""
        )

    def is_candidate(self) -> bool:
        """
        Return True if this version is a candidate version.
        """
        return self.candidate is not None

    def bumped_major(self, other: "Version") -> bool:
        """
        Return True if this version is a major bump from the other version.
        """
        return self.major > other.major

    def bumped_functional(self, other: "Version") -> bool:
        """
        Return True if this version is a functional bump from the other version.
        Return False if major bump is detected.
        """
        return not self.bumped_major(other) and self.functional > other.functional

    def bumped_technical(self, other: "Version") -> bool:
        """
        Return True if this version is a technical bump from the other version.
        Return False if major or functional bump is detected.
        """
        return not self.bumped_functional(other) and self.technical > other.technical

    def bumped_candidate(self, other: "Version") -> bool:
        """
        Return True if this version is a candidate bump from the other version.
        Return False if major, functional or technical bump is detected.
        Raises ValueError if one of the versions is not a candidate version.
        """
        if self.candidate is None or other.candidate is None:
            raise ValueError("Cannot compare candidate versions if one of them is not a candidate.")
        return not self.bumped_technical(other) and self.candidate > other.candidate

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Version):
            return super().__eq__(other)
        if isinstance(other, str):
            return str(self) == other
        return NotImplemented

    def __lt__(self, other: "Version") -> bool:
        """
        This method asks: Is this (self) version older than the other version?
        """
        if not isinstance(other, Version):
            return NotImplemented
        for attr in ["major", "functional", "technical"]:
            if getattr(self, attr) != getattr(other, attr):
                return getattr(self, attr) < getattr(other, attr)
        if self.candidate != other.candidate:
            return self.candidate is not None and (other.candidate is None or self.candidate < other.candidate)
        return False  # self == other

    def __str__(self) -> str:
        return self.tag_name


def get_source_repo(gh_token: str | None = None) -> Repository:
    """
    Get the BO4E-python repository from GitHub.
    """
    if gh_token is not None:
        gh = Github(auth=Token(gh_token))
    else:
        gh = Github()
    return gh.get_repo("bo4e/BO4E-python")


def get_latest_version(gh_token: str | None = None) -> Version:
    """
    Get the release from BO4E-python repository which is marked as 'latest'.
    """
    return Version.from_string(get_source_repo(gh_token).get_latest_release().tag_name)


def is_version_tag(value: str) -> bool:
    """
    Check if value is a valid version tag and exists in repository.
    """
    try:
        Version.from_string(value, allow_candidate=True)
        subprocess.check_call(["git", "show-ref", "--quiet", f"refs/tags/{value}"])
    except (ValueError, subprocess.CalledProcessError):
        return False
    return True


def is_branch(value: str) -> bool:
    """
    Check if a branch is a valid branch name and exists in repository.
    """
    try:
        subprocess.check_call(["git", "show-ref", "--quiet", f"refs/remotes/origin/{value}"])
        return True
    except subprocess.CalledProcessError:
        return False


def get_branches_containing_commit(commit_id: str) -> Iterable[str]:
    """
    Get all branches containing the commit id.
    If the commit id is not found, a subprocess.CalledProcessError will be raised.
    If the commit exists but is not on any branch (e.g. only on tags), an empty Iterable will be returned.
    """
    cmd = ["git", "branch", "-a", "--contains", commit_id]
    output = subprocess.check_output(cmd).decode().strip()
    if output.startswith("error: no such commit"):
        raise subprocess.CalledProcessError(1, cmd, output=output)
    return (line.strip().lstrip("*").lstrip() for line in output.splitlines())


def is_commit(value: str) -> bool:
    """
    Check if value is a valid commit id.
    """
    try:
        if re.fullmatch(r"^[0-9a-f]{40}$", value) is None:
            return False
        _ = get_branches_containing_commit(value)
        # If the commit ID doesn't exist, an error will be raised.
    except subprocess.CalledProcessError:
        return False
    return True


def get_checkout_commit_id() -> str:
    """
    Get the commit id of the current checkout.
    """
    return subprocess.check_output(["git", "rev-parse", "HEAD"]).decode().strip()


def _get_ref(ref: str) -> tuple[Literal["tag", "branch", "commit"], str]:
    """
    Get the type of reference and the reference itself.
    """
    if is_version_tag(ref):
        logger.info("Get tags before tag %s", ref)
        return "tag", ref
    if is_branch(ref):
        logger.info("Get tags on branch %s", ref)
        return "branch", ref
    if is_commit(ref):
        logger.info("Get tags before commit %s", ref)
        return "commit", ref
    cur_commit = get_checkout_commit_id()
    logger.info(
        "Supplied value (%s) is neither a tag, a branch nor a commit. Get tags before current checkout commit %s",
        ref,
        cur_commit,
    )
    return "commit", cur_commit


def get_last_n_tags(
    n: int, *, ref: str = "main", exclude_candidates: bool = True, exclude_technical_bumps: bool = False
) -> Iterable[str]:
    """
    Get the last n tags in chronological descending order starting from `ref`.
    If `ref` is a branch, it will start from the current HEAD of the branch.
    If `ref` is a tag, it will start from the tag itself. But the tag itself will not be included in the output.
    If `ref` is neither nor, the main branch will be used as fallback.
    If `exclude_candidates` is True, candidate versions will be excluded from the output.
    If the number of found versions is less than `n`, a warning will be logged.
    If n=0, all versions since v202401.0.0 will be taken into account.
    If exclude_technical_bumps is True, from each functional release group,
    the highest technical release will be returned.
    """
    version_threshold = "v202401.0.0"  # Is used if n=0
    ref_type, reference = _get_ref(ref)
    if n == 0:
        logger.info("Get all tags since %s", version_threshold)
    else:
        logger.info("Get the last %d tags", n)

    logger.info("%s release candidates", "Exclude" if exclude_candidates else "Include")
    logger.info("%s technical bumps", "Exclude" if exclude_technical_bumps else "Include")
    output = (
        subprocess.check_output(["git", "tag", "--merged", reference, "--sort=-creatordate"])
        .decode()
        .strip()
        .splitlines()
    )
    if len(output) == 0:
        logger.warning("No tags found.")
        return
    last_version = Version.from_string(output[0], allow_candidate=True)

    counter = 0
    stop_iteration = False
    for ind, tag in enumerate(output):
        if counter >= n > 0:
            stop_iteration = True
        if stop_iteration:
            return
        if n == 0 and tag == version_threshold:
            stop_iteration = True
        version = Version.from_string(tag, allow_candidate=True)
        # pylint: disable=too-many-boolean-expressions
        if (
            exclude_candidates
            and version.is_candidate()
            or exclude_technical_bumps
            and ind > 0
            and not last_version.bumped_functional(version)
            and not last_version.bumped_major(version)
            or ind == 0
            and ref_type == "tag"
        ):
            logger.info("Skipping version %s", version)
            continue
        logger.info("Yielding version %s", version)
        yield tag
        last_version = version
        counter += 1
    if counter < n and 0 < n:
        if ref_type == "tag":
            logger.warning("Only found %d tags before tag %s, tried to retrieve %d", counter, ref, n)
        else:
            logger.warning("Only found %d tags on branch %s, tried to retrieve %d", counter, ref, n)
    if n == 0:
        logger.warning("Threshold version %s not found. Returned all tags.", version_threshold)


def get_last_version_before(version: Version) -> Version:
    """
    Get the last non-candidate version before the provided version following the commit history.
    """
    return Version.from_string(one(get_last_n_tags(1, ref=version.tag_name)))


def ensure_latest_on_main(latest_version: Version, is_cur_version_latest: bool) -> None:
    """
    Ensure that the latest release is on the main branch.
    Will also be called if the currently tagged version is marked as `latest`.
    In this case both versions are equal.

    Note: This doesn't revert the release on GitHub. If you accidentally released on the wrong branch, you have to
    manually mark an old or create a new release as `latest` on the main branch. Otherwise, the publish workflow
    will fail here.
    """
    commit_id = subprocess.check_output(["git", "rev-parse", f"tags/{latest_version.tag_name}~0"]).decode().strip()
    branches_containing_commit = get_branches_containing_commit(commit_id)
    if "remotes/origin/main" not in branches_containing_commit:
        if is_cur_version_latest:
            raise ValueError(
                f"Tagged version {latest_version} is marked as latest but is not on main branch "
                f"(branches {branches_containing_commit} contain commit {commit_id}).\n"
                "Either tag on main branch or don't mark the release as latest.\n"
                "If you accidentally marked the release as latest please remember to revert it. "
                "Otherwise, the next publish workflow will fail as the latest version is assumed to be on main."
            )
        raise ValueError(
            f"Fatal Error: Latest release {latest_version.tag_name} is not on main branch "
            f"(branches {branches_containing_commit} contain commit {commit_id}).\n"
            "Please ensure that the latest release is on the main branch."
        )


def compare_work_tree_with_latest_version(
    gh_version: str, gh_token: str | None = None, major_bump_allowed: bool = True
) -> None:
    """
    Compare the work tree with the latest release from the BO4E repository.
    If any inconsistency is detected, a Value- or an AssertionError will be raised.
    """
    logger.info("Github Access Token %s", "provided" if gh_token is not None else "not provided")
    cur_version = Version.from_string(gh_version, allow_candidate=True)
    logger.info("Tagged release version: %s", cur_version)
    latest_version = get_latest_version(gh_token)
    logger.info("Got latest release version from GitHub: %s", latest_version)
    is_cur_version_latest = cur_version == latest_version
    if is_cur_version_latest:
        logger.info("Tagged version is marked as latest.")
    ensure_latest_on_main(latest_version, is_cur_version_latest)
    logger.info("Latest release is on main branch.")

    version_ahead = cur_version
    version_behind = get_last_version_before(cur_version)
    logger.info(
        "Comparing with the version before the tagged release (excluding release candidates): %s",
        version_behind,
    )

    assert version_ahead > version_behind, f"Version did not increase: {version_ahead} <= {version_behind}"

    logger.info(
        "Current version is ahead of the compared version. Comparing versions: %s -> %s",
        version_behind,
        version_ahead,
    )
    if version_ahead.bumped_major(version_behind):
        if not major_bump_allowed:
            raise ValueError("Major bump detected. Major bump is not allowed.")
        logger.info("Major version bump detected. No further checks needed.")
        return
    changes = list(
        diff.compare_bo4e_versions(version_behind.tag_name, version_ahead.tag_name, gh_token=gh_token, from_local=True)
    )
    logger.info("Check if functional or technical release bump is needed")
    functional_changes = len(changes) > 0
    logger.info("%s release bump is needed", "Functional" if functional_changes else "Technical")

    if not functional_changes and version_ahead.bumped_functional(version_behind):
        raise ValueError(
            "Functional version bump detected but no functional changes found. "
            "Please bump the technical release count instead of the functional."
        )
    if functional_changes and not version_ahead.bumped_functional(version_behind):
        raise ValueError(
            "No functional version bump detected but functional changes found. "
            "Please bump the functional release count.\n"
            f"Detected changes: {changes}"
        )


@click.command()
@click.option("--gh-version", type=str, required=True, help="The new version to compare the latest release with.")
@click.option(
    "--gh-token", type=str, default=None, help="GitHub Access token. This helps to avoid rate limiting errors."
)
@click.option(
    "--major-bump-allowed/--major-bump-disallowed",
    is_flag=True,
    default=True,
    help="Indicate if a major bump is allowed. "
    "If it is not allowed, the script will exit with an error if a major bump is detected.",
)
def compare_work_tree_with_latest_version_cli(
    gh_version: str, gh_token: str | None = None, major_bump_allowed: bool = True
) -> None:
    """
    Check a version tag and compare the work tree with the latest release from the BO4E repository.
    Exits with status code 1 iff the version is inconsistent with the commit history or if the detected changes in
    the JSON-schemas are inconsistent with the version bump.
    """
    try:
        compare_work_tree_with_latest_version(gh_version, gh_token, major_bump_allowed)
    except Exception as error:
        logger.error("An error occurred.", exc_info=error)
        raise click.exceptions.Exit(1)
    logger.info("All checks passed.")


if __name__ == "__main__":
    # pylint: disable=no-value-for-parameter
    compare_work_tree_with_latest_version_cli()


def test_compare_work_tree_with_latest_version() -> None:
    """
    Little test function for local testing.
    """
    logging.basicConfig(level=logging.DEBUG, stream=sys.stdout)
    compare_work_tree_with_latest_version("v202401.1.2-rc3", gh_token=None)


def test_version() -> None:
    """
    Test the total ordering of the Version class.
    """
    # pylint: disable=unnecessary-negation
    assert Version.from_string("v202401.1.2") == Version(major=202401, functional=1, technical=2)
    assert Version.from_string("v202401.1.2-rc3", allow_candidate=True) == Version(
        major=202401, functional=1, technical=2, candidate=3
    )
    assert Version.from_string("v202401.1.2") < Version.from_string("v202401.1.3")
    assert Version.from_string("v202401.1.2") < Version.from_string("v202401.2.0")
    assert not Version.from_string("v202401.2.0") < Version.from_string("v202401.1.2")
    assert Version.from_string("v202401.2.0") > Version.from_string("v202401.1.2")
    assert Version.from_string("v202401.1.2-rc3", allow_candidate=True) < Version.from_string("v202401.1.2")
    assert Version.from_string("v202401.1.2-rc3", allow_candidate=True) <= Version.from_string("v202401.1.2")
    assert not Version.from_string("v202401.1.2-rc3", allow_candidate=True) >= Version.from_string("v202401.1.2")
    assert Version.from_string("v202401.1.2-rc3", allow_candidate=True) > Version.from_string("v202401.1.1")
    assert Version.from_string("v202401.1.2-rc3", allow_candidate=True) > Version.from_string(
        "v202401.1.2-rc1", allow_candidate=True
    )
    assert Version.from_string("v202501.2.0") > Version.from_string("v202401.10.23")



"""
Contains the logic to detect the different changes between two BO4E versions.
"""

import itertools
import logging
import re
from pathlib import Path
from typing import Iterable, Sequence

from bost.schema import AllOf, AnyOf, Array, Object, Reference, SchemaRootType, SchemaType, StrEnum, String, TypeBase

from . import change_schemas, loader

logger = logging.getLogger(__name__)

REGEX_IGNORE_VERSION = re.compile(r"v\d+\.\d+\.\d+(-rc\d+)?")


def _diff_type_base(
    schema_old: TypeBase, schema_new: TypeBase, old_trace: str, new_trace: str
) -> Iterable[change_schemas.Change]:
    """
    This function compares two type base schemas and yields the changes.
    """
    if schema_old.title != schema_new.title:
        logger.warning(
            (
                "Title should not change. Renaming is not detectable and the titles are autogenerated.\n"
                "'%s' -> '%s'\n"
                "%s -> %s"
            ),
            schema_old.title,
            schema_new.title,
            old_trace,
            new_trace,
        )
        yield change_schemas.Change(
            type=change_schemas.ChangeType.FIELD_TITLE_CHANGED,
            old=schema_old.title,
            new=schema_new.title,
            old_trace=old_trace,
            new_trace=new_trace,
        )
    if REGEX_IGNORE_VERSION.sub(schema_old.description, "{__gh_version__}") != REGEX_IGNORE_VERSION.sub(
        schema_new.description, "{__gh_version__}"
    ):
        yield change_schemas.Change(
            type=change_schemas.ChangeType.FIELD_DESCRIPTION_CHANGED,
            old=schema_old.description,
            new=schema_new.description,
            old_trace=old_trace,
            new_trace=new_trace,
        )
    if schema_old.default != schema_new.default and schema_old.title != " Version" and schema_new.title != " Version":
        yield change_schemas.Change(
            type=change_schemas.ChangeType.FIELD_DEFAULT_CHANGED,
            old=schema_old.default,
            new=schema_new.default,
            old_trace=old_trace,
            new_trace=new_trace,
        )


def _diff_enum_schemas(
    schema_old: StrEnum, schema_new: StrEnum, old_trace: str, new_trace: str
) -> Iterable[change_schemas.Change]:
    """
    This function compares two enum schemas and yields the changes.
    """
    new_enum_values = set(schema_new.enum)
    for old_enum_value in schema_old.enum:
        if old_enum_value not in new_enum_values:
            yield change_schemas.Change(
                type=change_schemas.ChangeType.ENUM_VALUE_REMOVED,
                old=old_enum_value,
                new=None,
                old_trace=old_trace,
                new_trace=new_trace,
            )
        else:
            new_enum_values.remove(old_enum_value)
    for new_enum_value in new_enum_values:
        yield change_schemas.Change(
            type=change_schemas.ChangeType.ENUM_VALUE_ADDED,
            old=None,
            new=new_enum_value,
            old_trace=old_trace,
            new_trace=new_trace,
        )


def _diff_object_schemas(
    schema_old: Object, schema_new: Object, old_trace: str, new_trace: str
) -> Iterable[change_schemas.Change]:
    """
    This function compares two object schemas and yields the changes.
    """
    new_properties = set(schema_new.properties.keys())
    for key, value in schema_old.properties.items():
        if key not in schema_new.properties:
            yield change_schemas.Change(
                type=change_schemas.ChangeType.FIELD_REMOVED,
                old=value,
                new=None,
                old_trace=f"{old_trace}.properties['{key}']",
                new_trace=new_trace,
            )
        else:
            new_properties.remove(key)

            # Field exists in both schemas, check for further changes
            yield from _diff_schema_type(
                value,
                schema_new.properties[key],
                f"{old_trace}.properties['{key}']",
                f"{new_trace}.properties['{key}']",
            )
    for key in new_properties:
        yield change_schemas.Change(
            type=change_schemas.ChangeType.FIELD_ADDED,
            old=None,
            new=schema_new.properties[key],
            old_trace=old_trace,
            new_trace=f"{new_trace}.properties['{key}']",
        )


def _diff_ref_schemas(
    schema_old: Reference, schema_new: Reference, old_trace: str, new_trace: str
) -> Iterable[change_schemas.Change]:
    """
    This function compares two reference schemas and yields a change if the references are different.
    Even if the referenced schema only got renamed or moved, the reference will be treated as different
    because in any client application you would have to update the references.
    """
    if schema_old.ref != schema_new.ref:
        yield change_schemas.Change(
            type=change_schemas.ChangeType.FIELD_REFERENCE_CHANGED,
            old=schema_old.ref,
            new=schema_new.ref,
            old_trace=old_trace,
            new_trace=new_trace,
        )


def _diff_array_schemas(
    schema_old: Array, schema_new: Array, old_trace: str, new_trace: str
) -> Iterable[change_schemas.Change]:
    """
    This function compares two array schemas and yields the changes.
    """
    yield from _diff_schema_type(schema_old.items, schema_new.items, f"{old_trace}.items", f"{new_trace}.items")


def _diff_any_of_or_all_of_schemas(
    schema_old: AnyOf | AllOf, schema_new: AnyOf | AllOf, old_trace: str, new_trace: str
) -> Iterable[change_schemas.Change]:
    """
    This function compares two anyOf or allOf schemas and yields the changes.
    """
    assert type(schema_old) is type(schema_new), "Internal error: This function should only be called for equal types"
    if isinstance(schema_old, AnyOf):
        query_str = "any_of"
    else:
        query_str = "all_of"
    found_new_types = set()
    for old_index, old_type in enumerate(getattr(schema_old, query_str)):
        found_old_in_new = False
        for new_index, new_type in enumerate(getattr(schema_new, query_str)):
            changes = list(
                _diff_schema_type(
                    old_type, new_type, f"{old_trace}.{query_str}[{old_index}]", f"{new_trace}.{query_str}[{new_index}]"
                )
            )
            if not any(change_schemas.filter_non_crit(changes)):
                # The types are equal (except for non-critical changes), yield the non-critical changes
                found_old_in_new = True
                assert new_index not in found_new_types, "Internal error: Duplicate type in anyOf"
                found_new_types.add(new_index)
                yield from changes
                break
        if not found_old_in_new:
            yield change_schemas.Change(
                type=(
                    change_schemas.ChangeType.FIELD_ANY_OF_TYPE_REMOVED
                    if isinstance(schema_old, AnyOf)
                    else change_schemas.ChangeType.FIELD_ALL_OF_TYPE_REMOVED
                ),
                old=old_type,
                new=None,
                old_trace=f"{old_trace}.{query_str}[{old_index}]",
                new_trace=f"{new_trace}.{query_str}",
            )
    not_found_indices = set(range(len(getattr(schema_new, query_str)))) - found_new_types
    for new_index in not_found_indices:
        yield change_schemas.Change(
            type=(
                change_schemas.ChangeType.FIELD_ANY_OF_TYPE_ADDED
                if isinstance(schema_old, AnyOf)
                else change_schemas.ChangeType.FIELD_ALL_OF_TYPE_ADDED
            ),
            old=None,
            new=getattr(schema_new, query_str)[new_index],
            old_trace=old_trace,
            new_trace=f"{new_trace}.{query_str}[{new_index}]",
        )


def _diff_string_schemas(
    schema_old: String, schema_new: String, old_trace: str, new_trace: str
) -> Iterable[change_schemas.Change]:
    """
    This function compares two string schemas and yields the changes.
    """
    if schema_old.format != schema_new.format:
        yield change_schemas.Change(
            type=change_schemas.ChangeType.FIELD_STRING_FORMAT_CHANGED,
            old=schema_old.format,
            new=schema_new.format,
            old_trace=old_trace,
            new_trace=new_trace,
        )


def _diff_schema_differing_types(
    schema_old: SchemaType, schema_new: SchemaType, old_trace: str, new_trace: str
) -> Iterable[change_schemas.Change]:
    """
    This function compares two differing schema types and yields the changes.
    """
    assert type(schema_old) is not type(
        schema_new
    ), "Internal error: This function should only be called for differing types"
    # Types are different. Check if it is "only" a change in cardinality
    if isinstance(schema_old, Object) and isinstance(schema_new, Array):
        sub_changes = list(_diff_schema_type(schema_old, schema_new.items, old_trace, f"{new_trace}.items"))
    elif isinstance(schema_old, Array) and isinstance(schema_new, Object):
        sub_changes = list(_diff_schema_type(schema_old.items, schema_new, f"{old_trace}.items", new_trace))
    else:
        sub_changes = None

    if sub_changes is None or any(change_schemas.filter_non_crit(sub_changes)):
        # Treat the types as equal iff there are no critical changes between the types
        # In if-Block, the types are different
        yield change_schemas.Change(
            type=change_schemas.ChangeType.FIELD_TYPE_CHANGED,
            old=schema_old,
            new=schema_new,
            old_trace=old_trace,
            new_trace=new_trace,
        )
    else:
        # If the types are equal (except for non-critical changes), yield the non-critical changes
        # plus a change in cardinality
        yield from sub_changes
        # If the type of one schema is equal to the items type of the other, there is a change in cardinality
        yield change_schemas.Change(
            type=change_schemas.ChangeType.FIELD_CARDINALITY_CHANGED,
            old=schema_old,
            new=schema_new,
            old_trace=old_trace,
            new_trace=new_trace,
        )


def _diff_schema_type(
    schema_old: SchemaType, schema_new: SchemaType, old_trace: str, new_trace: str
) -> Iterable[change_schemas.Change]:
    """
    This function compares two schema types and yields the changes.
    """
    yield from _diff_type_base(schema_old, schema_new, old_trace, new_trace)
    if type(schema_old) is not type(schema_new):
        yield from _diff_schema_differing_types(schema_old, schema_new, old_trace, new_trace)
    # Even if the types are equal on this shallow level, we must do some more checks for certain
    # types.
    elif isinstance(schema_new, StrEnum):
        yield from _diff_enum_schemas(schema_old, schema_new, old_trace, new_trace)  # type: ignore[arg-type]
        # mypy isn't able to know that type(schema_new) is type(schema_old) here (and in the following)
    elif isinstance(schema_new, Object):
        yield from _diff_object_schemas(schema_old, schema_new, old_trace, new_trace)  # type: ignore[arg-type]
    elif isinstance(schema_new, Reference):
        yield from _diff_ref_schemas(schema_old, schema_new, old_trace, new_trace)  # type: ignore[arg-type]
    elif isinstance(schema_new, Array):
        yield from _diff_array_schemas(schema_old, schema_new, old_trace, new_trace)  # type: ignore[arg-type]
    elif isinstance(schema_new, (AnyOf, AllOf)):
        yield from _diff_any_of_or_all_of_schemas(
            schema_old,  # type: ignore[arg-type]
            schema_new,
            old_trace,
            new_trace,
        )
    # Any other types are definitely equal at this point


def _diff_root_schemas(
    schema_old: SchemaRootType, schema_new: SchemaRootType, old_trace: str, new_trace: str
) -> Iterable[change_schemas.Change]:
    """
    This function compares two root schemas and yields the changes.
    """
    yield from _diff_schema_type(schema_old, schema_new, old_trace, new_trace)


def diff_schemas(
    schemas_old: Path, schemas_new: Path, old_trace: str, new_trace: str
) -> Iterable[change_schemas.Change]:
    """
    This function compares two BO4E versions and yields the changes.
    Note: The paths to the old and the new schemas should correspond to the same root node of the tree structure.
    I.e. the direct subdirectories should be "bo", "com" and "enum".
    """
    old_schema_files = {file.relative_to(schemas_old) for file in schemas_old.rglob("*.json")}
    new_schema_files = {file.relative_to(schemas_new) for file in schemas_new.rglob("*.json")}

    for schema_file in old_schema_files - new_schema_files:
        yield change_schemas.Change(
            type=change_schemas.ChangeType.CLASS_REMOVED,
            old=loader.load_schema_file(schemas_old / schema_file),
            new=None,
            old_trace=f"{old_trace}/{'/'.join(schema_file.with_suffix('').parts)}#",
            new_trace=f"{new_trace}/#",
        )
    for schema_file in new_schema_files - old_schema_files:
        yield change_schemas.Change(
            type=change_schemas.ChangeType.CLASS_ADDED,
            old=None,
            new=loader.load_schema_file(schemas_new / schema_file),
            old_trace=f"{old_trace}/#",
            new_trace=f"{new_trace}/{'/'.join(schema_file.with_suffix('').parts)}#",
        )
    for schema_file in old_schema_files & new_schema_files:
        yield from _diff_root_schemas(
            loader.load_schema_file(schemas_old / schema_file),
            loader.load_schema_file(schemas_new / schema_file),
            f"{old_trace}/{'/'.join(schema_file.with_suffix('').parts)}#",
            f"{new_trace}/{'/'.join(schema_file.with_suffix('').parts)}#",
        )


def compare_bo4e_versions(
    version_old: str, version_new: str, gh_token: str | None = None, from_local: bool = False
) -> Iterable[change_schemas.Change]:
    """
    Compare the old version with the new version.
    If version_new is None use the BO4E version of the checkout working directory by assuming the compiled json
    schemas in /json_schemas.
    """
    dir_old_schemas = loader.pull_or_reuse_bo4e_version(version_old, gh_token)
    dir_new_schemas = loader.pull_or_reuse_bo4e_version(version_new, gh_token, from_local=from_local)
    print(f"Comparing {version_old} with {version_new}")
    yield from diff_schemas(dir_old_schemas, dir_new_schemas, version_old, version_new)


def compare_bo4e_versions_iteratively(
    versions: Sequence[str], cur_version: str | None = None, gh_token: str | None = None
) -> dict[tuple[str, str], Iterable[change_schemas.Change]]:
    """
    Compare the versions iteratively. Each version at index i will be compared to the version at index i+1.
    Additionally, if cur_version is provided, the last version in the list will be compared to the version
    in the checkout working directory. The value of cur_version will be used to set the key in the returned
    dict.
    Note:
        - versions must contain at least one element.
        - versions should be sorted in ascending order.
        - if using cur_version, ensure that the json schemas of the checkout working directory
          were build on beforehand. They should be located in /json_schemas.
    """
    print(f"Comparing versions {versions} with cur_version {cur_version}")
    if len(versions) == 0:
        print("No versions to compare.")
        return {}
    changes = {}
    last_version: str = versions[0]  # This value is never used but makes mypy and pylint happy
    for version_old, version_new in itertools.pairwise(versions):
        last_version = version_new
        changes[version_old, version_new] = compare_bo4e_versions(version_old, version_new, gh_token)
    if cur_version is not None:
        changes[last_version, cur_version] = compare_bo4e_versions(last_version, cur_version, gh_token, from_local=True)
    print("Comparisons finished.")
    return changes
